
----------------------- Local Data -----------------------

-- Options table
local options = {
	type='group',
	args = {
		config = {
			name = "config",
			desc = "Configure operational options",
			type = "group",
			args = {
				health = {
					name = "health",
					type = 'toggle',
					desc = "Display player health with combat messages",
					get = function () return Expiration.db.profile.health; end,
					set = function () Expiration.db.profile.health = not Expiration.db.profile.health; end,
				},
				event = {
					name = "event",
					type = 'toggle',
					desc = "Include the combat event with the logged message",
					get = function () return Expiration.db.profile.event; end,
					set = function () Expiration.db.profile.event = not Expiration.db.profile.event; end,
				},
				lines = {
					name = "lines",
					type = 'range',
					desc = "Number of lines to track per player",
					min = 5,
					max = 50,
					step = 1,
					get = function () return Expiration.db.profile.lines; end,
		  			set = function (value) Expiration.db.profile.lines = value; end,
				},
				reports = {
					name = "reports",
					type = 'range',
					desc = "Number of reports to track per player",
					min = 1,
					max = 20,
					step = 1,
					get = function () return Expiration.db.profile.reports; end,
		  			set = function (value) Expiration.db.profile.reports = value; end,
				},
				threshold = {
					name = "threshold",
					type = 'range',
					desc = "Don't log messages with damage/healing values below this amount",
					min = 0,
					max = 5000,
					step = 100,
					get = function () return Expiration.db.profile.threshold; end,
					set = function (value) Expiration.db.profile.threshold = value; end,
				},
				dest = {
					name = "dest",
					type = 'text',
					desc = "Default report channel",
					get = function () return Expiration.db.profile.dest; end,
					set = function (text) Expiration.db.profile.dest = text; end,
					validate = {"console", "party", "raid", "guild", "officer"},
				},
			},
		},
		refresh = {
			name = "refresh",
			type = 'execute',
			desc = "Refresh the tracking of party/raid units",
			func = "UpdateUnits",
		},
		report = {
			name = "report",
			type = 'text',
			input = true,
			desc = "Show a report on a given player",
			get = false,
			set = "Report",
			usage = "<name> <report# or 'last'> [dest[:target]] [lines]",
		},
		deaths = {
			name = "deaths",
			type = 'text',
			input = true,
			desc = "List the available reports in the last specified interval",
			get = false,
			set = "ListDeaths",
			usage = "<minutes>",
		},
	},
};

-- Events to track
-- table value is location of amount parameter, if event should be subject to thresholding
local events = {
	ENVIRONMENTAL_DAMAGE = 2,
	SWING_DAMAGE = 1,
	-- SWING_MISSED = -1,
	RANGE_DAMAGE = 4,
	-- RANGE_MISSED = -1,
	-- SPELL_CAST_START = -1,
	-- SPELL_CAST_SUCCESS = -1,
	-- SPELL_CAST_FAILED = -1,
	-- SPELL_MISSED = -1,
	SPELL_DAMAGE = 4,
	SPELL_HEAL = 4,
	-- SPELL_ENERGIZE = 4,
	SPELL_DRAIN = 4,
	SPELL_LEECH = 4,
	-- SPELL_SUMMON = -1,
	-- SPELL_CREATE = -1,
	SPELL_INSTAKILL = -1,
	-- SPELL_INTERRUPT = -1,
	-- SPELL_EXTRA_ATTACKS = -1,
	-- SPELL_DURABILITY_DAMAGE = -1,
	-- SPELL_DURABILITY_DAMAGE_ALL = -1,
	SPELL_AURA_APPLIED = -1,
	SPELL_AURA_APPLIED_DOSE = -1,
	SPELL_AURA_REMOVED = -1,
	SPELL_AURA_REMOVED_DOSE = -1,
	SPELL_AURA_DISPELLED = -1,
	SPELL_AURA_STOLEN = -1,
	-- ENCHANT_APPLIED = -1,
	-- ENCHANT_REMOVED = -1,
	-- SPELL_PERIODIC_MISSED = -1,
	SPELL_PERIODIC_DAMAGE = 4,
	SPELL_PERIODIC_HEAL = 4,
	-- SPELL_PERIODIC_ENERGIZE = 4,
	SPELL_PERIODIC_DRAIN = 4,
	SPELL_PERIODIC_LEECH = 4,
	SPELL_DISPEL_FAILED = -1,
	DAMAGE_SHIELD = 4,
	-- DAMAGE_SHIELD_MISSED = -1,
	-- DAMAGE_SPLIT = 4,
	-- PARTY_KILL = -1,
	UNIT_DIED = -1,
	UNIT_DESTROYED = -1,
};

-- Events that can kill people
local killerevents = {
	ENVIRONMENTAL_DAMAGE = true,
	SWING_DAMAGE = true,
	RANGE_DAMAGE = true,
	SPELL_DAMAGE = true,
	SPELL_DRAIN = true,
	SPELL_LEECH = true,
	SPELL_INSTAKILL = true,
	SPELL_PERIODIC_DAMAGE = true,
	SPELL_PERIODIC_DRAIN = true,
	SPELL_PERIODIC_LEECH = true,
	DAMAGE_SHIELD = true,
};

-- Translate icons id's into icon tags
local icons = {"{star}", "{circle}", "{diamond}", "{triangle}", "{moon}", "{square}", "{cross}", "{skull}"};
local function getIconTag(id) return icons[tonumber(id)]; end

-- Common global functions
local pairs, ipairs = pairs, ipairs;
local tinsert, tremove = tinsert, tremove;
local format, strlower, gsub = format, strlower, gsub;

-- Debug flag
local debug = false;

----------------------- AddOn Functions -----------------------

-- Instantiate mod
Expiration = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0");
local Expiration = Expiration;

Expiration:RegisterChatCommand({"/expiration", "/exp"}, options);
Expiration:RegisterDB("ExpirationDB");
Expiration:RegisterDefaults("profile", {
		health = true,
		lines = 20,
		reports = 5,
		dest = "console",
		threshold = 0,
	});
Expiration.units = { };

-- Unit subtable format:
-- Expiration.units[UnitGUID(unit)] = {
--		name = UnitName(unit),
--		unit = unit,
--		dead = true/nil,			(whether or not unit is dead)
--		lastevent = "..."/nil,		(last recorded event type)
--		lastsource = "..."/nil,		(source of last recorded event)
--		history = { ... },
--		reports = {
--			{
--				time = time(),
--				cause = "Environment"/sourceName/"???",
--				...
--			},
--			...
--		},
--	}


-- Enable the mod
function Expiration:OnEnable()
	
	-- Register Party/Raid Events
	self:RegisterBucketEvent({"PLAYER_ENTERING_WORLD", "PARTY_MEMBERS_CHANGED", "RAID_ROSTER_UPDATE"}, 1, "UpdateUnits");
	
	-- Register Combat Events
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "CombatEvent");
	self:RegisterEvent("UNIT_HEALTH", function (unit) Expiration:DeathCheck(UnitGUID(unit), time()); end);
	
	-- Update unit table
	self:UpdateUnits();
	
end


-- Strip tags from strings that are not safe to send with SendChatMessage
function Expiration:CleanString(message)
	
	-- strip color codes
	message = gsub(message, "|c%x%x%x%x%x%x%x%x(.-)|r", "%1");
	
	-- strip unit codes
	message = gsub(message, "|Hunit:0x%x+:(.-)|h.-|h", "%1");
	
	-- strip action codes
	message = gsub(message, "|Haction:.-|h(.-)|h", "%1");
	
	-- convert icon file codes to {icon} tags
	message = gsub(message, "|Hicon:%d+:%a-|h|TInterface\\TargetingFrame\\UI%-RaidTargetingIcon_(%d)%.blp:%d|t|h", getIconTag);
	
	-- strip spell links
	message = gsub(message, "|Hspell:.-:.-|h(.-)|h", "%1");
	
	-- check for final |'s
	if debug and strfind(message, "|") then
		self:Print("Unstripped tag found in " .. gsub(message, "|", "||"));
	end
	
	return message;
	
end


-- Fill out the list of players to track appropriately
function Expiration:UpdateUnits()
	
	-- set up new unit table
	local new_units = { };
	
	-- function to add a unit to be tracked
	local trackUnit =
		function (unit)
			
			if not UnitExists(unit) then return; end
			
			local id = UnitGUID(unit);
			if self.units[id] then
				new_units[id] = self.units[id];
				new_units[id].unit = unit;
				return;
			end
			
			new_units[id] = {
				name = UnitName(unit),
				unit = unit,
				history = { },
				reports = { },
			};
			
		end
	
	-- Player is in a raid
	if GetNumRaidMembers() ~= 0 then
		
		-- rebuild list of raid members
		for i = 1, 40 do
			trackUnit("raid"..i);
			trackUnit("raid"..i.."pet");
		end
		
	-- Player is in a party
	elseif GetNumPartyMembers() ~= 0 then
		
		-- rebuild list of party members
		for i = 1, 4 do
			trackUnit("party"..i);
			trackUnit("party"..i.."pet");
		end
		
	end
	
	-- Add the player
	trackUnit("player");
	trackUnit("pet");
	
	-- clean out any event data that doesn't correspond to someone in the raid
	self.units = new_units;
	
end


-- Checks whether the specified unit is dead (true if dead, false if not)
function Expiration:DeathCheck(id, timestamp)
	
	local unit = self.units[id];
	
	if not unit then return false; end
	
	-- check unit death status
	if ( UnitHealth(unit.unit) == 0 ) or UnitIsDeadOrGhost(unit.unit) then
		
		-- unit just died, record death event and make report
		if not unit.dead then
			unit.dead = true;
			self:AddEvent(id, timestamp, format(UNITDIESOTHER, unit.name));
			self:AddReport(id);
		end
		
		return true;
		
	else
		
		-- clear death status
		if unit.dead then
			unit.dead = nil;
			unit.lastevent = nil;
			unit.lastsource = nil;
		end
		
	end
	
	return false;
	
end


-- Handle a combat event
function Expiration:CombatEvent(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	
	-- check event argument
	if not event then return; end
	
	-- get local copy of unit table
	local unit = self.units[destGUID];
	
	-- check destGUID
	if not unit then return; end
	
	-- check unit death status
	if self:DeathCheck(destGUID, timestamp) then return; end
	
	-- check event table
	if not events[event] then return; end
	
	-- if potential killer event, record in unit
	if killerevents[event] then
		unit.lastevent = event;
		unit.lastsource = sourceName;
	end
	
	-- check threshold filter
	if events[event] > 0 then
		local amount = select(events[event], ...);
		if ( amount < self.db.profile.threshold ) then return; end
	end
	
	-- get text string (TODO: set up own filter settings)
	local message = CombatLog_OnEvent(Blizzard_CombatLog_CurrentSettings, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...);
	
	-- append event string to end
	if self.db.profile.event then message = format("%s (%s)", message, event); end
	
	-- add event
	self:AddEvent(destGUID, timestamp, message);
	
end


-- Add an event message to the specified id history
function Expiration:AddEvent(id, timestamp, message)
	
	if not id or not self.units[id] then return; end
	if not message then return; end
	
	-- get local copy of unit table
	local unit = self.units[id];
	
	-- remove oldest message to make room if necessary
	if #unit.history >= self.db.profile.lines then
		tremove(unit.history, 1);
	end
	
	-- format message
	local msg = format("%05.2f ", timestamp % 60);
	
	if self.db.profile.health then
		local health = UnitHealth(unit.unit);
		local pcthp = health/UnitHealthMax(unit.unit);
		local red, green;
		
		if ( pcthp > 0.5 ) then
			red = 510*(1 - pcthp);
			green = 255;
		else
			red = 255;
			green = 510*pcthp;
		end
		
		msg = msg..format("[|cff%02X%02X00%d %d%%|r] ", red, green, health, 100*pcthp);
	end
	
	msg = msg..message;
	
	tinsert(unit.history, msg);
	
end


-- copy the current unit history as a death report
function Expiration:AddReport(id)
	
	if not id or not self.units[id] then return; end
	
	local unit = self.units[id];
	
	-- create report table
	local report = { };
	
	report.time = time();
	report.cause = ( unit.lastevent == "ENVIRONMENTAL_DAMAGE" ) and "Environment" or unit.lastsource or "???";
	
	if debug and ( report.cause == "???" ) then
		self:Print("Unknown cause of death for unit " .. unit.name .. " with last event: " .. (unit.lastevent or "nil") .. " and last source: " .. (unit.lastsource or "nil"));
	end
	
	for i, msg in ipairs(unit.history) do report[i] = msg; end
	
	-- add report
	if #unit.reports >= self.db.profile.reports then
		tremove(unit.reports, 1);
	end
	
	tinsert(unit.reports, report);
	
end


-- Send a report
function Expiration:Report(name, report, dest, lines)
	
	-- check arguments
	if not name or ( type(name) ~= "string" ) then
		self:Print("Enter a player or pet to report on.");
		return;
	end
	
	if report and ( type(report) ~= "number" ) and ( strlower(report) ~= "last" ) then
		self:Print("Report must be a number or 'last'.");
		return;
	end
	
	if dest and ( type(dest) ~= "string" ) then
		self:Print("Destination must be a string.");
		return;
	end
	
	if lines and ( type(lines) ~= "number" ) then
		self:Print("Lines must be a number.");
		return;
	end
	
	
	-- find the unit name
	local id = nil;
	
	for guid, unit in pairs(self.units) do
		if strlower(name) == strlower(unit.name) then
			id = guid; break;
		end
	end
	
	-- unit not found
	if not id then
		self:Print("No unit found with name \'"..name.."\'");
		return;
	end
	
	-- local reference to unit table
	local unit = self.units[id];
	
	-- if no report specified, list reports
	if not report then
		
		if ( #unit.reports < 1 ) then
			self:Print("No reports available for " .. unit.name);
			return;
		end
		
		self:Print("Reports available for " .. unit.name .. ":");
		
		for i, report in ipairs(unit.reports) do
			self:Print(i .. ". " .. date("%H:%M:%S", report.time) .. " - " .. report.cause);
		end
		
		return;
		
	end
	
	-- accept 'last' as a report designation
	if ( strlower(report) == "last" ) then report = #unit.reports; end
	
	if not unit.reports[report] then
		self:Print("Report " .. report .. " for " .. unit.name .. " not found.");
		return;
	end
	
	-- local reference to report
	report = unit.reports[report];
	
	-- use default destination if none provided
	if not dest then dest = self.db.profile.dest; end
	
	-- set up args
	if type(dest) == "number" then
		lines = dest;
		dest = self.db.profile.dest;
	else
		dest = strlower(dest);
	end
	
	-- set up lines to report
	if not lines then lines = self.db.profile.lines; end
	lines = (( lines < 1 ) and 1) or (( lines > #report ) and #report) or lines;
	
	-- Set print function based on output type
	local print;
	
	-- separate type from to
	local type, target = strsplit(":", dest, 2);
	
	local chatType = {
		party = "PARTY",
		raid = "RAID",
		guild = "GUILD",
		officer = "OFFICER",
		whisper = "WHISPER",
		channel = "CHANNEL",
		
		p = "PARTY",
		r = "RAID",
		g = "GUILD",
		o = "OFFICER",
		w = "WHISPER",
		c = "CHANNEL",
	};
	
	-- print to console
	if ( type == "console" ) then
		print = function (text) self:Print(text); end;
		
	-- print to a channel
	elseif chatType[type] then
		
		-- convert
		type = chatType[type];
		
		-- catch chat failures
		if ( type == "PARTY" ) and ( GetNumPartyMembers() == 0 ) then
			SendChatMessage("placeholder", "PARTY");
			return;
			
		elseif ( type == "RAID" ) and ( GetNumRaidMembers() == 0 ) then
			SendChatMessage("placeholder", "RAID");
			return;
			
		elseif (( type == "GUILD" ) or ( type == "OFFICER" )) and not IsInGuild() then
			SendChatMessage("placeholder", "GUILD");
			return;
			
		elseif ( type == "WHISPER" ) then
			
			if not target or ( strtrim(target) == "" ) then
				self:Print("Whisper destinations must have a target player.");
				return;
			end
			
		elseif ( type == "CHANNEL" ) then
			
			if not target or ( strtrim(target) == "" ) then
				self:Print("Channel destinations must have a target channel.");
				return;
			end
			
			local cid = GetChannelName(target);
			
			if not cid or ( cid == 0 ) then
				self:Print("No such channel: \'" .. target .. "\'");
				return;
			else
				target = cid;
			end
			
		end
		
		print = function (text) SendChatMessage(self:CleanString(text), type, nil, target); end;
		
	-- report error
	else
		self:Print("Unknown destination type \'" .. type .. "\'");
		return;
	end
	
	-- Print report
	print("Expiration report for "..unit.name.." (Last "..lines.." events)");
	
	for line = #report - lines + 1, #report do
		print((#report - line + 1)..". "..report[line]);
	end
	
end


-- List available death reports for the last n minutes
function Expiration:ListDeaths(minutes)
	
	-- check argument
	if not minutes then return; end
	if type(minutes) ~= "number" then
		self:Print("Minutes must be a number.");
		return;
	end
	
	-- set time frame
	local start = time() - 60*minutes;
	
	-- create death reports table
	local reports = { };
	
	-- for each unit
	for guid, unit in pairs(self.units) do
		-- for each report
		for id, report in ipairs(unit.reports) do
			-- if within timeframe, add summary table to reports
			if ( report.time > start ) then
				tinsert(reports, {id = id, name = unit.name, cause = report.cause, time = report.time});
			end	
		end
	end
	
	-- sort summary table
	sort(reports, function (a, b) return a.time < b.time; end);
	
	-- list reports
	self:Print(#reports.." reports available from the last "..minutes.." minutes");
	
	for i, report in ipairs(reports) do
		self:Print(format("%s - %s #%d - %s", date("%H:%M:%S", report.time), report.name or "nil", report.id or -1, report.cause or "nil"));	
	end
	
end
